import importlib.util
import os
import sys
from pint import UnitRegistry
from pathlib import Path

import warnings


def check_qt_presence():
    try:
        import qtpy

    except ImportError as e:
        msg = f"\n\n" \
              f"****************************************************************************************\n" \
              f"No Qt backend could be found in your system, please install either pyqt5/6 or pyside2/6.\n\n" \
              f"pyqt5 is still preferred, while pyqt6 should mostly work.\n\n" \
              f"do:\n" \
              f"pip install pyqt5\n for instance\n"\
              f"****************************************************************************************\n"
        warnings.warn(msg, FutureWarning, 2)
        sys.exit()


check_qt_presence()


try:
    from pymodaq_utils.logger import set_logger
    from pymodaq_utils.utils import get_version, PackageNotFoundError
    try:
        __version__ = get_version(__package__)
    except PackageNotFoundError:
        __version__ = '0.0.0dev'
    try:
        logger = set_logger('pymodaq_gui', base_logger=False)
    except Exception:
        print("Couldn't create the local folder to store logs , presets...")

    logger.info('')
    logger.info('')
    logger.info('************************')
    logger.info('Starting PyMoDAQ modules')
    logger.info('************************')
    logger.info('')
    logger.info('')


    # in a try statement for compilation on readthedocs server but if this fail, you cannot use the code
    from pymodaq_gui.plotting import data_viewers  # imported here as to avoid circular imports later on
    from pymodaq_gui.qt_utils import setLocale, set_qt_backend

    from pymodaq_utils.config import Config
    from pymodaq_data.plotting.plotter.plotter import register_plotter, PlotterFactory

    # issue on windows when using .NET code within multithreads, this below allows it but requires the
    # pywin32 (pythoncom) package
    if importlib.util.find_spec('clr') is not None:
        try:
            import pythoncom
            pythoncom.CoInitialize()
        except ModuleNotFoundError as e:
            infos = "You have installed plugins requiring the pywin32 package to work correctly," \
                    " please type in *pip install pywin32* and restart PyMoDAQ"
            print(infos)
            logger.warning(infos)

    config = Config()  # to ckeck for config file existence, otherwise create one

    logger.info('************************')
    logger.info(f"Setting Qt backend to: {config['qtbackend']['backend']} ...")
    set_qt_backend()
    logger.info('************************')
    logger.info('')
    logger.info('')
    logger.info('************************')
    logger.info(f"Setting Locale to {config['style']['language']} / {config['style']['country']}")
    logger.info('************************')
    setLocale()
    logger.info('')
    logger.info('')

    logger.info('')
    logger.info('')
    logger.info('************************')
    logger.info(f"Registering PyMoDAQ qt plotters...")
    register_plotter(parent_module_name='pymodaq_gui.plotting.plotter')
    logger.info(f"Done")
    logger.info('************************')

except Exception as e:
    try:
        logger.exception(str(e))
    except Exception as e:
        print(str(e))



from qtpy.QtCore import QObject
from qtpy import QtWidgets


class ListPicker(QObject):

    def __init__(self, list_str):
        super(ListPicker, self).__init__()
        self.list = list_str

    def pick_dialog(self):
        self.dialog = QtWidgets.QDialog()
        self.dialog.setMinimumWidth(500)
        vlayout = QtWidgets.QVBoxLayout()

        self.list_widget = QtWidgets.QListWidget()
        self.list_widget.addItems(self.list)

        vlayout.addWidget(self.list_widget, 10)
        self.dialog.setLayout(vlayout)

        buttonBox = QtWidgets.QDialogButtonBox()
        buttonBox.addButton('Apply', buttonBox.AcceptRole)
        buttonBox.accepted.connect(self.dialog.accept)
        buttonBox.addButton('Cancel', buttonBox.RejectRole)
        buttonBox.rejected.connect(self.dialog.reject)

        vlayout.addWidget(buttonBox)
        self.dialog.setWindowTitle('Select an entry in the list')

        res = self.dialog.show()

        pass
        if res == self.dialog.Accepted:
            # save managers parameters in a xml file
            return [self.list_widget.currentIndex(), self.list_widget.currentItem().text()]
        else:
            return [-1, ""]


import numpy as np
from pyqtgraph.parametertree import Parameter
from qtpy import QtCore, QtWidgets, QtGui
from qtpy.QtCore import QPointF, Slot, Signal, QObject
from typing import List, Tuple

from pyqtgraph import LinearRegionItem

from pymodaq_utils.logger import set_logger, get_module_name
from pymodaq_utils import utils
from pymodaq_utils import math_utils as mutils

from pymodaq_data.data import DataFromRoi, DataToExport, Axis, DataWithAxes
from pymodaq_data import data as data_mod
from pymodaq_data.post_treatment.process_to_scalar import DataProcessorFactory

from pymodaq_gui.managers.roi_manager import ROIManager, LinearROI, RectROI
from pymodaq_gui.plotting.items.crosshair import Crosshair
from pymodaq_gui.plotting.items.image import UniformImageItem
from pymodaq_gui.plotting.data_viewers.viewer1Dbasic import Viewer1DBasic


logger = set_logger(get_module_name(__file__))

data_processors = DataProcessorFactory()


class Filter:

    def __init__(self):
        self._is_active = False
        self._slot_to_send_data = None

    def register_activation_signal(self, activation_signal):
        activation_signal.connect(lambda x: self.set_active(x))

    def register_target_slot(self, slot):
        self._slot_to_send_data = slot

    @Slot(bool)
    def set_active(self, activate=True):
        self._is_active = activate

    def filter_data(self, data: data_mod.DataRaw):
        if self._is_active:
            filtered_data = self._filter_data(data)
            if filtered_data is not None and self._slot_to_send_data is not None:
                self._slot_to_send_data(filtered_data)

    def _filter_data(self, data: data_mod.DataRaw) -> DataToExport:
        raise NotImplementedError


class Filter1DFromCrosshair(Filter):
    def __init__(self, crosshair: Crosshair):
        """
        Extract data along a crosshair using coordinates and data displayed in graph_items such as  imageItems
        Parameters
        ----------
        crosshair : Crosshair
        """
        super().__init__()
        self.crosshair = crosshair
        self._x, self._y = 0., 0.
        self._axis: data_mod.Axis = None

    def update_axis(self, axis: data_mod.Axis):
        self._axis = axis

    def _filter_data(self, data: data_mod.DataRaw) -> DataToExport:
        dte = DataToExport('Crosshair')
        if data is not None:
            axis = data.get_axis_from_index(0, create=False)[0]
            if axis is not None:
                self.update_axis(axis)

                self._x, self._y = self.crosshair.get_positions()
                dwa = data.isig[data.axes[0].find_indexes([self._x])[0]]
                dwa.axes = [Axis('x', data=np.array([self._x]))]
                dte.append(dwa)
                # for label, dat in zip(data.labels, data.data):
                # dte.append(DataFromRoi('crosshair', data=[np.array([dat[ind_x]]) for dat in data.data],
                #                        axes=[Axis(data=np.array([self._axis.get_data()[ind_x]]))],
                #                        labels=data.labels))
        return dte


class Filter2DFromCrosshair(Filter):
    def __init__(self, crosshair: Crosshair, graph_items, image_keys):
        """
        Extract data along a crosshair using coordinates and data displayed in graph_items such as  imageItems
        Parameters
        ----------
        crosshair : (Crosshair)
        graph_items : (dict)
        image_keys : (list) list of string identifier to link datas to their graph_items. This means that in
            _filter_data, datas['data'][key] is plotted on graph_items[key] for key in image_keys
        """
        super().__init__()
        self._graph_items = graph_items
        self._image_keys = image_keys
        self.crosshair = crosshair
        self._x, self._y = 0., 0.

    def set_graph_items(self, graph_items):
        self._graph_items = graph_items

    @Slot(bool)
    def set_active(self, activate=True):
        self._is_active = activate
        if activate:
            self.crosshair.crosshair_dragged.emit(*self.crosshair.get_positions())

    def _filter_data(self, dwa: data_mod.DataRaw) -> DataToExport:
        dte = DataToExport('Crosshair')
        if dwa is not None:
            self._x, self._y = self.crosshair.get_positions()
            data_type = dwa.distribution
            if data_type == 'uniform':
                dte = self.get_data_from_uniform(dwa)
            elif data_type == 'spread':
                dte = self.get_data_from_spread(dwa)
        return dte

    def get_data_from_uniform(self, dwa: DataWithAxes) -> DataToExport:
        indx, indy = self.mapfromview(self._x, self._y, 'red')

        data_H_index = slice(None, None, 1)
        data_V_index = slice(None, None, 1)
        H_indexes = (utils.rint(indy), data_H_index)
        V_indexes = (data_V_index, utils.rint(indx))
        dte = DataToExport('Crosshair')
        try:
            if not (0 <= utils.rint(indy) < dwa.shape[0]):
                raise IndexError
            dwa_hor = dwa.isig[H_indexes]
            dwa_hor.labels = [f'Crosshair/{label}' for label in dwa_hor.labels]
            dwa_hor.name = 'hor'
            dte.append(dwa_hor)
        except IndexError:
            pass
        try:
            if not (0 <= utils.rint(indx) < dwa.shape[1]):
                raise IndexError
            dwa_ver = dwa.isig[V_indexes]
            dwa_ver.labels = [f'Crosshair/{label}' for label in dwa_ver.labels]
            dwa_ver.name = 'ver'
            dte.append(dwa_ver)
        except IndexError:
            pass
        try:
            if not (0 <= utils.rint(indy) < dwa.shape[0]) \
                or \
                    not (0 <= utils.rint(indx) < dwa.shape[1]):
                raise IndexError
            dwa_int = dwa.isig[utils.rint(indy), utils.rint(indx)]
            dwa_int.labels = [f'Crosshair/{label}' for label in dwa_int.labels]
            dwa_int.name = 'int'
            dte.append(dwa_int)
        except IndexError:
            pass
        return dte

    def get_data_from_spread(self, dwa: DataWithAxes) -> DataToExport:

        data_H_index = slice(None, None, 1)
        data_V_index = slice(None, None, 1)
        posx, posy = self.mapfromview(self._x, self._y, 'red')

        hor_data = []
        ver_data = []
        int_data = []
        hor_axis = None
        ver_axis = None

        for ind, data_key in enumerate(self._graph_items):
            if ind < len(dwa):
                points, data = self._graph_items[data_key].get_points_at(axis='y', val=posy)
                x_sorted_indexes = np.argsort(points[:, 0])
                hor_axis = points[x_sorted_indexes, 0][data_H_index]

                hor_data.append(data[x_sorted_indexes][data_H_index])

                points, data = self._graph_items[data_key].get_points_at(axis='x', val=posx)
                y_sorted_indexes = np.argsort(points[:, 1])
                ver_axis = points[y_sorted_indexes, 1][data_V_index]

                ver_data.append(data[y_sorted_indexes][data_V_index])

                int_data.append(np.array([self._graph_items[data_key].get_val_at((posx, posy))]))

        dte = DataToExport('Crosshair')
        if len(hor_data) > 0 and len(hor_axis) > 0:
            dte.append(DataFromRoi('hor', data=hor_data,
                                   axes=[Axis(dwa.axes[1].label, dwa.axes[1].units, data=hor_axis)]),)
        if len(ver_data) > 0 and len(ver_axis) > 0:
            dte.append(DataFromRoi('ver', data=ver_data,
                                   axes=[Axis(dwa.axes[0].label, dwa.axes[0].units, data=ver_axis)]))
        if len(int_data) > 0:
            dte.append(DataFromRoi('int', data=int_data))

        return dte

    def mapfromview(self, x, y, item_key='red'):
        """
        get item coordinates from view coordinates
        Parameters
        ----------
        x: (float) x coordinate in the view reference frame
        y: (float) y coordinate in the view refernece frame

        Returns
        -------
        x: (float) coordinate in the item reference frame
        y: (float) coordinate in the item reference frame
        """
        point = self._graph_items[item_key].mapFromView(QPointF(x, y))
        return point.x(), point.y()


class Filter1DFromRois(Filter):
    """

    Parameters
    ----------
    roi_manager:ROIManager
    graph_item: PlotItems
    """
    def __init__(self, roi_manager: ROIManager):

        super().__init__()
        self._roi_settings = roi_manager.settings
        self._ROIs = roi_manager.ROIs
        self._axis: data_mod.Axis = None

    def update_axis(self, axis: data_mod.Axis):
        self._axis = axis

    def _filter_data(self, data: data_mod.DataRaw) -> DataToExport:
        dte = DataToExport('roi1D')
        try:
            axis = data.get_axis_from_index(0, create=False)[0]
            if axis is not None:
                self.update_axis(axis)
            if data is not None:
                for roi_key, roi in self._ROIs.items():
                    if self._roi_settings['ROIs', roi_key, 'use_channel'] == 'All':
                        data_index = list(range(len(data.labels)))
                    else:
                        try:
                            data_index = [data.labels.index(self._roi_settings['ROIs', roi_key,
                                          'use_channel'])]
                        except ValueError:
                            data_index = [0]
                    dte_tmp = self.get_data_from_roi(roi, self._roi_settings.child('ROIs', roi_key),
                                                     data)
                    if self._roi_settings['ROIs', roi_key, 'use_channel'] == 'All':
                        dte.append(dte_tmp.data)
                    else:
                        for index in data_index:
                            for dwa in dte_tmp.data:
                                dte.append(dwa.pop(index))

        except Exception as e:
            pass
        finally:
            return dte

    def get_data_from_roi(self, roi: LinearROI,  roi_param: Parameter, data: data_mod.DataWithAxes) -> DataToExport:
        if data is not None:
            dte = DataToExport('ROI1D')
            _slice = self.get_slice_from_roi(roi, data)
            sub_data: DataFromRoi = data.isig[_slice]
            sub_data.name = 'HorData'
            sub_data.origin = roi_param.name()
            sub_data.labels = [f'{roi_param.name()}/{label}' for label in sub_data.labels]
            dte.append(sub_data)
            if sub_data.size != 0:
                processed_data = data_processors.get(roi_param['math_function']).process(sub_data)
            else:
                processed_data = None
            if processed_data is not None:
                processed_data.name = 'IntData'
                dte.append(processed_data)
            return dte

    def get_slice_from_roi(self, roi: RectROI, data: data_mod.DataWithAxes) -> slice:
        ind_x_min, ind_x_max = data.get_axis_from_index(data.sig_indexes[0])[0].find_indexes(roi.getRegion())
        size = data.get_axis_from_index(0)[0].size
        ind_x_min = int(min(max(ind_x_min, 0), size))
        ind_x_max = int(max(0, min(ind_x_max, size)))
        return slice(ind_x_min, ind_x_max)


class Filter2DFromRois(Filter):
    """Filters 2D data using 2D ROIs

    Parameters
    ----------
    roi_manager: ROIManager
    graph_item: UniformImageItem or SpreadImageItem
        The graphical item where data and ROIs are plotted
    image_keys : (list) list of string identifier to link datas to their graph_items. This means that in
        _filter_data, datas.data[key] is plotted on graph_items[key] for key in image_keys
    """
    def __init__(self, roi_manager: ROIManager, graph_item: UniformImageItem, image_keys):

        super().__init__()
        self._roi_settings = roi_manager.settings
        self._image_keys = image_keys
        self._graph_item = graph_item
        self.axes = (0, 1)
        self._ROIs = roi_manager.ROIs

    def _filter_data(self, dwa: data_mod.DataRaw) -> DataToExport:
        dte = DataToExport('ROI')
        if dwa is not None:
            try:
                labels = []
                for roi_key, roi in self._ROIs.items():
                    label = self._roi_settings['ROIs', roi_key, 'use_channel']
                    if label is not None:
                        if label != 'All':
                            sub_data = dwa.deepcopy()
                            sub_data.data = [dwa[dwa.labels.index(label)]]
                            sub_data.labels = [label]
                        else:
                            sub_data = dwa
                        dte_temp = self.get_xydata_from_roi(roi, sub_data,
                                                            self._roi_settings['ROIs',
                                                            roi_key, 'math_function'])

                        dte.append(dte_temp)
            except Exception as e:
                logger.warning(f'Issue with the ROI: {str(e)}')
        return dte

    def get_slices_from_roi(self, roi: RectROI, data_shape: tuple) -> Tuple[slice, slice]:
        x, y = roi.pos().x(), roi.pos().y()
        width, height = roi.size().x(), roi.size().y()
        size_y, size_x = data_shape
        ind_x_min = int(min(max(x, 0), size_x))
        ind_y_min = int(min(max(y, 0), size_y))
        ind_x_max = int(max(0, min(x+width, size_x)))
        ind_y_max = int(max(0, min(y+height, size_y)))
        return slice(ind_y_min,ind_y_max), slice(ind_x_min, ind_x_max)

    def get_xydata_from_roi(self, roi: RectROI, dwa: DataWithAxes, math_function: str) -> DataToExport:
        dte = DataToExport(roi.name)
        if dwa is not None:
            labels = [f'{roi.name}/{label}' for label in dwa.labels]
            if dwa.distribution.name == 'spread':
                xvals, yvals, data = self.get_xydata_spread(dwa, roi)
                if len(data) == 0:
                    return dte
                ind_xaxis = np.argsort(xvals)
                ind_yaxis = np.argsort(yvals)
                xvals = xvals[ind_xaxis]
                yvals = yvals[ind_yaxis]
                data_H = data[ind_xaxis]
                data_V = data[ind_yaxis]
                int_data = np.array([np.mean(data)])

                _x_axis = dwa.get_axis_from_index_spread(0, 0)
                x_axis = Axis(_x_axis.label, _x_axis.units, data=xvals, index=0, spread_order=0)
                _y_axis = dwa.get_axis_from_index_spread(0, 1)
                y_axis = Axis(_y_axis.label, _y_axis.units, data=yvals, index=0, spread_order=0)
                sub_data_hor = DataFromRoi('hor', distribution='spread', data=[data_H], axes=[x_axis],)
                sub_data_ver = DataFromRoi('ver', distribution='spread', data=[data_V], axes=[y_axis])
                math_data = DataFromRoi('int', data=int_data)
            else:
                slices = self.get_slices_from_roi(roi, dwa.shape)
                sub_data: DataFromRoi = dwa.isig[slices[0], slices[1]]
                sub_data_hor = sub_data.mean(0)
                sub_data_ver = sub_data.mean(1)
                math_data = data_processors.get(math_function).process(sub_data)

            sub_data_hor.name = 'hor'
            sub_data_hor.origin = roi.name
            sub_data_hor.labels = labels
            sub_data_ver.name = 'ver'
            sub_data_ver.origin = roi.name
            sub_data_ver.labels = labels
            math_data.name = 'int'
            math_data.origin = roi.name
            math_data.labels = labels

            dte.append([sub_data_hor, sub_data_ver, math_data])
            return dte

    #TODO possibly not used anymore to be deleted
    #
    # def get_xydata(self, data: np.ndarray, roi: RectROI):
    #     data, coords = self.data_from_roi(data, roi)
    #
    #     if data is not None:
    #         xvals = np.linspace(np.min(np.min(coords[1, :, :])), np.max(np.max(coords[1, :, :])),
    #                             data.shape[1])
    #         yvals = np.linspace(np.min(np.min(coords[0, :, :])), np.max(np.max(coords[0, :, :])),
    #                             data.shape[0])
    #     else:
    #         xvals = yvals = data = np.array([])
    #     return xvals, yvals, data
    #
    # def data_from_roi(self, data, roi):
    #     data, coords = roi.getArrayRegion(data, self._graph_item, self.axes, returnMappedCoords=True)
    #     return data, coords

    def get_xydata_spread(self, data, roi):
        xvals = []
        yvals = []
        data_out = []
        for ind in range(data.shape[0]):
            # invoke the QPainterpath of the ROI (from the shape method)
            if roi.shape().contains(QPointF(data.get_axis_from_index(0)[0].get_data()[ind] - roi.pos().x(),
                                            data.get_axis_from_index(0)[1].get_data()[ind] - roi.pos().y())):
                xvals.append(data.get_axis_from_index(0)[0].get_data()[ind])
                yvals.append(data.get_axis_from_index(0)[1].get_data()[ind])
                data_out.append(data[0][ind])
        data_out = np.array(data_out)
        xvals = np.array(xvals)
        yvals = np.array(yvals)
        return xvals, yvals, data_out


class FourierFilterer(QObject):
    filter_changed = Signal(dict)

    def __init__(self, parent=None):
        super().__init__()
        if parent is None:
            parent = QtWidgets.QWidget()

        self.parent = parent

        self.raw_data = None
        self.data = None
        self.data_fft = None
        self.filter = None
        self.xaxis = None
        self.yaxis = None
        self.xaxisft = None
        self.yaxisft = None

        self.frequency = 0
        self.phase = 0

        self.c = None
        self.viewer2D = None
        self.setUI()

    def setUI(self):
        self.vlayout = QtWidgets.QVBoxLayout()
        self.parent.setLayout(self.vlayout)

        form = QtWidgets.QWidget()
        self.viewer1D = Viewer1DBasic(form)
        self.vlayout.addWidget(form)
        self.fftbutton1D = QtWidgets.QPushButton()
        self.fftbutton1D.setText("")
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/FFT.png"), QtGui.QIcon.Normal,
                       QtGui.QIcon.Off)
        self.fftbutton1D.setIcon(icon)
        self.fftbutton1D.setCheckable(True)
        self.fftbutton1D.clicked.connect(self.update_plot)

        vbox = self.viewer1D.parent.layout()
        widg = QtWidgets.QWidget()
        hbox = QtWidgets.QHBoxLayout()
        widg.setLayout(hbox)
        vbox.insertWidget(0, widg)
        hbox.addWidget(self.fftbutton1D)
        hbox.addStretch()

        self.viewer1D.ROI = LinearRegionItem(values=[0, 100])
        self.viewer1D.plotwidget.plotItem.addItem(self.viewer1D.ROI)
        self.data_filtered_plot = self.viewer1D.plotwidget.plotItem.plot()
        self.data_filtered_plot.setPen('w')
        self.viewer1D.ROI.sigRegionChangeFinished.connect(self.set_data)

        self.viewer1D.ROIfft = LinearRegionItem()
        self.viewer1D.plotwidget.plotItem.addItem(self.viewer1D.ROIfft)
        self.viewer1D.ROIfft.sigRegionChangeFinished.connect(self.update_filter)

        self.parent.show()

    def calculate_fft(self):

        ftaxis, axis = mutils.ftAxis_time(len(self.xaxis), np.max(self.xaxis) - np.min(self.xaxis))
        self.xaxisft = ftaxis / (2 * np.pi)
        self.data_fft = mutils.ft(self.data)

    def show_data(self, data):
        """
        show data and fft
        Parameters
        ----------
        data: (dict) with keys 'data', optionally 'xaxis' and 'yaxis'
        """
        try:
            self.raw_data = data

            if 'xaxis' in data:
                self.xaxis = data['xaxis']
            else:
                self.xaxis = np.arange(0, data['data'].shape[0], 1)
                self.raw_data['xaxis'] = self.xaxis
            # self.viewer1D.ROI.setRegion((np.min(self.xaxis), np.max(self.xaxis)))
            self.set_data()
        except Exception as e:
            logger.exception(str(e))

    def set_data(self):
        xlimits = self.viewer1D.ROI.getRegion()
        indexes = mutils.find_index(self.raw_data['xaxis'], xlimits)
        self.data = self.raw_data['data'][indexes[0][0]:indexes[1][0]]
        self.xaxis = self.raw_data['xaxis'][indexes[0][0]:indexes[1][0]]
        try:
            self.calculate_fft()
        except Exception as e:
            logger.exception(str(e))
        self.viewer1D.x_axis = self.xaxis
        self.update_plot()

    def update_filter(self):
        try:
            xmin, xmax = self.viewer1D.ROIfft.getRegion()
            self.filter = mutils.gauss1D(self.xaxisft, np.mean([xmin, xmax]), xmax - xmin)
            self.data = np.real(mutils.ift(self.filter * self.data_fft))
            index = np.argmax(self.filter * self.data_fft)
            self.frequency = self.xaxisft[index]
            self.phase = np.angle(self.data_fft[index])

            self.filter_changed.emit(dict(frequency=self.frequency, phase=self.phase))
            self.update_plot()
        except Exception as e:
            logger.exception(str(e))

    def update_plot(self):

        if self.fftbutton1D.isChecked():
            if self.data_fft is not None:
                if self.filter is not None:
                    self.viewer1D.show_data([np.abs(self.data_fft), np.max(np.abs(self.data_fft)) * self.filter])
                else:
                    self.viewer1D.show_data([np.abs(self.data_fft)])
                self.viewer1D.x_axis = dict(data=self.xaxisft, label='freq.')
                self.viewer1D.ROIfft.setVisible(True)
                self.viewer1D.ROI.setVisible(False)
                self.data_filtered_plot.setVisible(False)
        else:
            if self.raw_data is not None:
                self.viewer1D.show_data([self.raw_data['data']])
                self.viewer1D.x_axis = dict(data=self.raw_data['xaxis'], label='Pxls')
                self.data_filtered_plot.setData(self.xaxis, self.data)
                self.data_filtered_plot.setVisible(True)
                self.viewer1D.ROIfft.setVisible(False)
                self.viewer1D.ROI.setVisible(True)


if __name__ == '__main__':
    import sys

    app = QtWidgets.QApplication(sys.argv)
    prog = FourierFilterer()

    from pymodaq.utils.daq_utils import gauss1D

    xdata = np.linspace(0, 400, 401)
    x0 = 50
    dx = 20
    tau = 27
    tau2 = 100
    ydata_gauss = 10 * gauss1D(xdata, x0, dx) + np.random.rand(len(xdata))
    ydata_expodec = np.zeros((len(xdata)))
    ydata_expodec[:50] = 10 * gauss1D(xdata[:50], x0, dx, 2)
    ydata_expodec[50:] = 10 * np.exp(-(xdata[50:] - x0) / tau)  # +10*np.exp(-(xdata[50:]-x0)/tau2)
    ydata_expodec += 2 * np.random.rand(len(xdata))
    ydata_sin = 10 + 2 * np.sin(2 * np.pi * 0.1 * xdata - np.deg2rad(55)) + np.sin(
        2 * np.pi * 0.008 * xdata - np.deg2rad(-10)) + 2 * np.random.rand(len(xdata))

    prog.show_data(dict(data=ydata_sin, xaxis=xdata))
    sys.exit(app.exec_())


from qtpy import QtWidgets

from pyqtgraph.parametertree import parameterTypes, Parameter, ParameterTree
from . import pymodaq_ptypes


class ParameterTree(ParameterTree):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.header().setVisible(True)
        self.header().setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
        #self.header().setMinimumSectionSize(150)



import warnings

from qtpy.QtCore import Signal, QObject
from qtpy import QtGui, QtCore, QtWidgets
from pyqtgraph.dockarea import Dock, DockArea, DockLabel
from pyqtgraph.dockarea.DockArea import TempAreaWindow
from pyqtgraph.widgets.VerticalLabel import VerticalLabel


class Dock(Dock):
    dock_focused = Signal(str)
    def __init__(self, name, *args, fontSize='20px', **kwargs):
        kwargs['fontSize'] = fontSize
        super().__init__(name, *args, **kwargs)

    def removeWidgets(self):
        for widget in self.widgets:
            self.layout.removeWidget(widget)
            widget.close()
        self.widgets = []

    def setOrientation(self, o='auto', force=False):
        """
        Sets the orientation of the title bar for this Dock.
        Must be one of 'auto', 'horizontal', or 'vertical'.
        By default ('auto'), the orientation is determined
        based on the aspect ratio of the Dock.
        """
        if self.container() is not None:  # patch here for when Dock is closed and when the QApplication
            # event loop is processed
            if o == 'auto' and self.autoOrient:
                if self.container().type() == 'tab':
                    o = 'horizontal'
                elif self.width() > self.height() * 1.5:
                    o = 'vertical'
                else:
                    o = 'horizontal'
            if force or self.orientation != o:
                self.orientation = o
                self.label.setOrientation(o)
                self.updateStyle()


class DockArea(DockArea, QObject):
    """
    Custom Dockarea class subclassing from the standard DockArea class and QObject so it can emit a signal when docks
    are moved around (one subclassed method: moveDock)
    See Also
    --------
    pyqtgraph.dockarea
    """
    dock_signal = Signal()

    def __init__(self, parent=None, temporary=False, home=None):
        super(DockArea, self).__init__(parent, temporary, home)

    def moveDock(self, dock, position, neighbor):
        """
        Move an existing Dock to a new location.
        """
        # Moving to the edge of a tabbed dock causes a drop outside the tab box
        if position in ['left', 'right', 'top',
                        'bottom'] and neighbor is not None and neighbor.container() is not None and neighbor.container().type() == 'tab':
            neighbor = neighbor.container()
        self.addDock(dock, position, neighbor)
        self.dock_signal.emit()

    def addTempArea(self):
        if self.home is None:
            area = DockArea(temporary=True, home=self)
            self.tempAreas.append(area)
            win = TempAreaWindow(area)
            area.win = win
            win.show()
        else:
            area = self.home.addTempArea()
        # print "added temp area", area, area.window()
        return area




import os
import numpy as np
from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtWidgets import QLabel
from qtpy.QtCore import QObject, Signal
from pymodaq_gui.utils.utils import clickable


class QLED(QLabel):
    value_changed = Signal(bool)

    def __init__(self, parent=None, scale=1, readonly=True):
        QLabel.__init__(self, parent)
        # self.setText("")
        self.red_icon = QtGui.QPixmap(":/icons/Icon_Library/red_light.png")
        self.green_icon = QtGui.QPixmap(":/icons/Icon_Library/greenLight2.png")
        self.setPixmap(self.red_icon)
        self.state = False
        self.clickable = not readonly
        # set the possibility to click and control the state of the LED otherwise it behaves as an indicator
        clickable(self).connect(
            self.LED_Clicked)
        # clickable is a function importing a filter class to deal with mouse down event as a signal see GUI_utils
        self.setText("empty")
        self.setMaximumWidth(self.height())
        if scale != 1:
            self.scale(scale)
        self.set_as_false()

    def scale(self, scale):
        self.green_icon = self.green_icon.scaled(scale * self.green_icon.width(),
                                                 scale * self.green_icon.height())
        self.red_icon = self.red_icon.scaled(scale * self.red_icon.width(),
                                             scale * self.red_icon.height())

    def get_state(self):
        return self.state

    def set_as(self, state=True):

        if state:
            self.set_as_true()
        else:
            self.set_as_false()
        if state != self.state:
            self.value_changed.emit(state)

    def set_as_true(self):
        self.state = True
        self.setPixmap(self.green_icon)

    def set_as_false(self):
        self.state = False
        self.setPixmap(self.red_icon)

    def LED_Clicked(self):
        if self.clickable:
            if self.state:
                self.set_as_false()
            else:
                self.set_as_true()
            self.value_changed.emit(not self.state)



from qtpy import QtWidgets, QtGui, QtCore
from pyqtgraph.parametertree.parameterTypes.list import ListParameter, ListParameterItem
from pyqtgraph.parametertree.Parameter import ParameterItem


class Combo_pb(QtWidgets.QWidget):

    def __init__(self, items=[]):
        super(Combo_pb, self).__init__()
        self.items = items
        self.initUI()
        self.count = self.combo.count

    def initUI(self):
        """
            Init the User Interface.
        """
        self.hor_layout = QtWidgets.QHBoxLayout()
        self.combo = QtWidgets.QComboBox()
        self.combo.addItems(self.items)
        self.add_pb = QtWidgets.QPushButton()
        self.add_pb.setText("")
        icon3 = QtGui.QIcon()
        icon3.addPixmap(QtGui.QPixmap(":/icons/Icon_Library/Add2.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        self.add_pb.setIcon(icon3)
        self.hor_layout.addWidget(self.combo)
        self.hor_layout.addWidget(self.add_pb)
        self.hor_layout.setSpacing(0)
        self.hor_layout.setContentsMargins(0, 0, 0, 0)
        self.add_pb.setMaximumWidth(25)
        self.setLayout(self.hor_layout)
        self.currentText = self.combo.currentText
        self.setCurrentIndex = self.combo.setCurrentIndex
        self.clear = self.combo.clear
        self.addItem = self.combo.addItem
        self.findText = self.combo.findText


class ListParameterItem(ListParameterItem):
    """
        WidgetParameterItem subclass providing comboBox that lets the user select from a list of options.

    """

    def __init__(self, param, depth):
        super().__init__(param, depth)

    def makeWidget(self):
        """
            Make a widget from self parameter options, connected to the buttonClicked function.

            Returns
            -------
            w:widget
                the initialized widget

            See Also
            --------
            buttonClicked, limitsChanged,
        """
        opts = self.param.opts
        t = opts['type']
        w = Combo_pb()
        w.add_pb.clicked.connect(self.buttonClicked)
        w.setMaximumHeight(20)  # # set to match height of spin box and line edit
        if 'show_pb' in opts:
            w.add_pb.setVisible(opts['show_pb'])
        else:
            w.add_pb.setVisible(False)
        w.sigChanged = w.combo.currentIndexChanged
        w.value = self.value
        w.setValue = self.setValue
        self.widget = w  # # needs to be set before limits are changed
        self.limitsChanged(self.param, self.param.opts['limits'])
        if len(self.forward) > 0:
            self.setValue(self.param.value())
        return w

    def buttonClicked(self):
        """
            |
            | Append the self limits attributes an added parameter with string value.
            | Update parameter and call the limitschanged method to map the added parameter.

            See Also
            --------
            limitsChanged,
        """
        if isinstance(self.param.opts['limits'], list):
            text, ok = QtWidgets.QInputDialog.getText(None, "Enter a value to add to the parameter",
                                                      "String value:", QtWidgets.QLineEdit.Normal)
            if ok and not (text == ""):
                self.param.opts['limits'].append(text)
                self.limitsChanged(self.param, self.param.opts['limits'])
                self.param.setValue(text)
        elif isinstance(self.param.opts['limits'], dict):
            text, ok = QtWidgets.QInputDialog.getText(None, "Enter a text to add to the parameter",
                                                      "String value:", QtWidgets.QLineEdit.Normal)
            if ok and not (text == ""):

                value, ok = QtWidgets.QInputDialog.getInt(None, "Enter an integer value to add to the parameter",
                                                          "integer value:", QtWidgets.QLineEdit.Normal)
                if ok:
                    self.param.opts['limits'].update({text: value})
                    self.limitsChanged(self.param, self.param.opts['limits'])
                    self.param.setValue(text)

    def optsChanged(self, param, opts):
        """
            Called when any options are changed that are not name, value, default, or limits.

            =============== ================================== =======================================
            **Parameters**    **Type**                           **Description**
            *param*           instance of pyqtgraph parameter    The parameter to be checked
            *opts*            string list                        The option dictionnary to be checked
            =============== ================================== =======================================

            See Also
            --------
            optsChanged
        """
        super().optsChanged(param, opts)
        if 'show_pb' in opts:
            self.widget.add_pb.setVisible(opts['show_pb'])
        if 'enabled' in opts:
            self.widget.setEnabled(opts['enabled'])


class ListParameter(ListParameter):
    """
        =============== =======================================
        **Attributes**    **Type**
        *itemClass*       instance of ListParameterItem_custom
        *sigActivated*    instance of pyqt Signal
        =============== =======================================
    """
    itemClass = ListParameterItem
    sigActivated = QtCore.Signal(object)

    def __init__(self, **opts):
        super().__init__(**opts)

    def activate(self):
        """
            Emit the Activated signal.
        """
        self.sigActivated.emit(self)
        self.emitStateChanged('activated', None)





from collections import OrderedDict
import copy
import datetime
import numpy as np
import sys
from typing import Union, Iterable, List, Dict

from qtpy import QtCore, QtGui, QtWidgets
from qtpy.QtCore import QObject, Slot, Signal
import pyqtgraph as pg
from pyqtgraph.graphicsItems.GradientEditorItem import Gradients
from pyqtgraph import ROI as pgROI

from pymodaq_utils import utils
from pymodaq_utils.logger import set_logger, get_module_name

from pymodaq_data.data import (Axis, DataToExport, DataFromRoi, DataRaw,
                               DataDistribution, DataWithAxes)

from pymodaq_gui.managers.roi_manager import ROIManager, SimpleRectROI
from pymodaq_gui.managers.action_manager import ActionManager
from pymodaq_gui.plotting.widgets import ImageWidget
from pymodaq_gui.plotting.data_viewers.viewer import ViewerBase
from pymodaq_gui.plotting.data_viewers.viewer1D import Viewer1D
from pymodaq_gui.plotting.data_viewers.viewer0D import Viewer0D
from pymodaq_gui.plotting.items.image import UniformImageItem, SpreadImageItem
from pymodaq_gui.plotting.items.axis_scaled import AXIS_POSITIONS, AxisItem_Scaled
from pymodaq_gui.plotting.items.crosshair import Crosshair
from pymodaq_gui.plotting.utils.filter import Filter2DFromCrosshair, Filter2DFromRois
from pymodaq_gui.plotting.utils.plot_utils import make_dashed_pens, RoiInfo


logger = set_logger(get_module_name(__file__))

Gradients.update(OrderedDict([
    ('red', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (255, 0, 0, 255))], 'mode': 'rgb'}),
    ('green', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (0, 255, 0, 255))], 'mode': 'rgb'}),
    ('blue', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (0, 0, 255, 255))], 'mode': 'rgb'}),
    ('spread', {'ticks': [(0.0, (0, 0, 0, 255)), (1.0, (255, 255, 255, 255))], 'mode': 'rgb'}),]))

COLORS_DICT = dict(red=(255, 0, 0), green=(0, 255, 0), blue=(0, 0, 255), spread=(128, 128, 128))



IMAGE_TYPES = ['red', 'green', 'blue']
COLOR_LIST = utils.plot_colors
crosshair_pens = make_dashed_pens(color=(255, 255, 0))


def image_item_factory(item_type='uniform', axisOrder='row-major', pen='r'):
    if item_type == 'uniform':
        image = UniformImageItem(pen=pen)
        image.setOpts(axisOrder=axisOrder)
    elif item_type == 'spread':
        image = SpreadImageItem(pen=pen)
    image.setCompositionMode(QtGui.QPainter.CompositionMode_Plus)
    return image


def histogram_factory(image_item=None, gradient='red'):
    """
    Create a pyqtgraph HistogramLUTWidget widget (histogram) and link it to the corresponding image_item
    Parameters
    ----------
    image_item: (ImageItem) the image item to be linked with the histogram
    gradient: (str) either 'red', 'green', 'blue', 'spread' or one of the Gradients

    Returns
    -------
    HistogramLUTWidget instance
    """

    if gradient not in Gradients:
        raise KeyError(f'Possible gradient are {Gradients} not {gradient}')

    histo = pg.HistogramLUTWidget()
    if image_item is not None:
        histo.setImageItem(image_item)

    histo.gradient.loadPreset(gradient)
    return histo


class ImageDisplayer(QObject):
    """
    This Object deals with the display of 2D data  with a plotitem
    """

    updated_item = Signal(dict)

    def __init__(self, plotitem, data_distribution: DataDistribution):
        super().__init__()
        self._plotitem = plotitem
        self._plotitem.addLegend()
        self.show_legend(False)
        self.display_type = data_distribution
        self._image_items = dict([])
        self._autolevels = False
        self._data: DataWithAxes = None

        self.update_display_items()

    def show_legend(self, show=True):
        self.legend.setVisible(show)

    @property
    def legend(self):
        return self._plotitem.legend

    def get_images(self):
        return self._image_items

    def get_image(self, name):
        if name not in self._image_items:
            raise KeyError(f'The image {name} is not defined in {self.__class__.__name__}')
        else:
            return self._image_items[name]

    @property
    def labels(self):
        if self._data is None:
            return []
        else:
            return self._data.labels

    @property
    def autolevels(self):
        return self._autolevels

    @Slot(bool)
    def set_autolevels(self, isautolevel):
        self._autolevels = isautolevel

    def update_data(self, dwa: DataWithAxes):
        if dwa.labels != self.labels:
            self.update_display_items(dwa.labels)
        if dwa.distribution != self.display_type:
            self.display_type = dwa.distribution
        self._data = dwa
        for ind_data, data_array in enumerate(dwa.data):
            if data_array.size > 0:
                if self.display_type == 'uniform':
                    self._image_items[IMAGE_TYPES[ind_data]].setImage(data_array, self.autolevels)
                else:
                    nav_axes = dwa.get_nav_axes()
                    data_array = np.stack((nav_axes[0].get_data(),
                                           nav_axes[1].get_data(),
                                           data_array), axis=0).T
                    self._image_items[IMAGE_TYPES[ind_data]].setImage(data_array, self.autolevels)

    def update_display_items(self, labels: List[str] = None):
        while len(self._image_items) > 0:
            self._plotitem.removeItem(self._image_items.pop(next(iter(self._image_items))))
        if labels is None:
            labels = []
            while len(labels) != len(IMAGE_TYPES):
                labels.append(IMAGE_TYPES[len(labels)])

        for ind, img_key in enumerate(IMAGE_TYPES):
            self._image_items[img_key] = image_item_factory(self.display_type, pen=img_key[0])
            self._plotitem.addItem(self._image_items[img_key])
            if ind < len(labels):
                self.legend.addItem(self._image_items[img_key], labels[ind])
        self.updated_item.emit(self._image_items)

    def update_image_visibility(self, are_items_visible):
        if len(are_items_visible) != len(self._image_items):
            raise ValueError(f'The length of the argument is not equal with the number of images')
        for ind, key in enumerate(IMAGE_TYPES):
            self._image_items[key].setVisible(are_items_visible[ind])


class Histogrammer(QObject):
    gradient_changed = Signal()

    def __init__(self, histogram_container: QtWidgets.QWidget, histogram_refs=IMAGE_TYPES):
        super().__init__()
        self._histograms = dict([])
        self._histogram_refs = histogram_refs
        self._histogram_container = histogram_container
        self.setup_histograms()
        self._autolevels = False

    def setup_histograms(self):
        for histo_key in self._histogram_refs:
            self._histograms[histo_key] = histogram_factory(None, gradient=histo_key)
            self.add_histogram(self._histograms[histo_key])
            self._histograms[histo_key].setVisible(False)
            self._histograms[histo_key].item.sigLookupTableChanged.connect(lambda: self.gradient_changed.emit())

    def get_histograms(self):
        return self._histograms

    def get_histogram(self, name):
        if name not in self.get_histograms():
            raise KeyError(f'The histogram {name} is not defined in {self.__class__.__name__}')
        else:
            return self._histograms[name]

    @property
    def autolevels(self):
        return self._autolevels

    @Slot(bool)
    def set_autolevels(self, isautolevels=True):
        self._autolevels = isautolevels
        for histo in self._histograms.values():
            histo.region.setVisible(not isautolevels)

    @Slot(bool)
    def activated(self, histo_action_checked):
        if histo_action_checked:
            for histo in self._histograms.values():
                histo.regionChanged()

    def affect_histo_to_imageitems(self, image_items):
        # TODO: if self._histogram_refs doesn't contains the same refs as image_items, we have an issue...
        for img_key in self._histogram_refs:
            self._histograms[img_key].setImageItem(image_items[img_key])

    def add_histogram(self, histogram):
        if self._histogram_container.layout() is None:
            self._histogram_container.setLayout(QtWidgets.QHBoxLayout())
        self._histogram_container.layout().addWidget(histogram)

    def show_hide_histogram(self, checked, are_items_visible):
        for ind_histo, histo_name in enumerate(self._histogram_refs):
            self._histograms[histo_name].setVisible(are_items_visible[ind_histo] and checked)

    def set_gradient(self, histo='red', gradient='grey'):
        """
        Change the color gradient of the specified histogram
        Parameters
        ----------
        histo: (str) either 'red', 'green', 'blue', 'spread' or 'all'
        gradient: (str or Gradient)
        """
        if gradient in Gradients:
            if histo == 'all':
                for key in self._histogram_refs:
                    self._histograms[key].item.gradient.loadPreset(gradient)
            else:
                self._histograms[histo].item.gradient.loadPreset(gradient)


class IsoCurver(QObject):
    def __init__(self, image_source, histogram_parent):
        super().__init__()
        self._histogram_parent = histogram_parent
        self.setup_iso_curve()
        self.update_image_source(image_source)
        self.update_histogram_parent(histogram_parent)
        self.show_hide_iso(False)

    def setup_iso_curve(self, parent_image_item='red'):
        # # Isocurve drawing
        self._isocurve_item = pg.IsocurveItem(level=0.8, pen='g', axisOrder='row-major')
        self._isocurve_item.setZValue(5)

        # # Draggable line for setting isocurve level
        self._isoLine = pg.InfiniteLine(angle=0, movable=True, pen='g')
        self._isoLine.setValue(0.8)
        self._isoLine.setZValue(1000)  # bring iso line above contrast controls

        self._isoLine.sigDragged.connect(self.update_isocurve)

    def update_image_source(self, image_source):
        self._isocurve_item.setParentItem(image_source)

    def update_histogram_parent(self, histogram):
        if histogram != self._histogram_parent:
            histogram.vb.removeItem(self._isoLine)
            self._histogram_parent = histogram
        histogram.vb.addItem(self._isoLine)
        histogram.vb.setMouseEnabled(y=False)  # makes user interaction a little easier

    def update_isocurve(self):
        self._isocurve_item.setLevel(self._isoLine.value())

    def set_isocurve_data(self, data):
        self._isocurve_item.setData(data)

    @Slot(bool)
    def show_hide_iso(self, show=True):
        if show:
            self._isocurve_item.show()
            self._isoLine.show()
        else:
            self._isocurve_item.hide()
            self._isoLine.hide()


class View2D(ActionManager, QtCore.QObject):

    lineout_types = ['hor', 'ver', 'int']

    def __init__(self, parent_widget=None):
        QtCore.QObject.__init__(self)
        ActionManager.__init__(self, toolbar=QtWidgets.QToolBar())

        self.ROIselect = SimpleRectROI([0, 0], [10, 10], centered=True, sideScalers=True)

        self._lineout_widgets = {widg_key: QtWidgets.QWidget() for widg_key in self.lineout_types}
        self.lineout_viewers: Dict[str, Viewer1D] = dict(hor=Viewer1D(self._lineout_widgets['hor'], show_toolbar=False, no_margins=True),
                                    ver=Viewer1D(self._lineout_widgets['ver'], show_toolbar=False, no_margins=True,
                                                 flip_axes=True),
                                    int=Viewer0D(self._lineout_widgets['int'], show_toolbar=False, no_margins=True))

        self.setup_actions()

        self.parent_widget = parent_widget
        if self.parent_widget is None:
            self.parent_widget = QtWidgets.QWidget()
            self.parent_widget.show()

        self.image_widget = ImageWidget()
        self.roi_manager = ROIManager(self.image_widget, '2D')

        self.roi_target: Union[pgROI, Crosshair] = None

        self.setup_widgets()

        self.histogrammer = Histogrammer(self.widget_histo)
        self.data_displayer: ImageDisplayer = None
        self.isocurver: IsoCurver = None

        self.crosshair = Crosshair(self.image_widget)

        self.connect_things()
        self.prepare_ui()

        self.set_axis_label('bottom', label='', units='Pxls')
        self.set_axis_label('left', label='', units='Pxls')

        self.set_image_displayer(DataDistribution['uniform'])

    def clear_plot_item(self):
        for item in self.plotitem.items[:]:
            if isinstance(item, SpreadImageItem) or isinstance(item, UniformImageItem):
                self.plotitem.removeItem(item)

    def set_image_displayer(self, data_distribution: DataDistribution):
        self.clear_plot_item()
        self.data_displayer = ImageDisplayer(self.plotitem, data_distribution)
        self.isocurver = IsoCurver(self.data_displayer.get_image('red'), self.histogrammer.get_histogram('red'))
        self.connect_action('isocurve', self.isocurver.show_hide_iso)
        self.data_displayer.updated_item.connect(self.histogrammer.affect_histo_to_imageitems)
        self.connect_action('autolevels', self.data_displayer.set_autolevels)
        for key in IMAGE_TYPES:
            self.connect_action(key, self.notify_visibility_data_displayer)

        self.histogrammer.affect_histo_to_imageitems(self.data_displayer.get_images())

        if data_distribution.name == 'uniform':
            self.roi_target = pgROI(pos=(0, 0), size=(20, 20), movable=False, rotatable=False,
                                    resizable=False)
            self.plotitem.addItem(self.roi_target)

        elif data_distribution.name == 'spread':
            self.roi_target = Crosshair(self.image_widget, pen=(255, 255, 255))
        self.roi_target.setVisible(False)

    def show_roi_target(self, show=True):
        self.roi_target.setVisible(show)

    def move_scale_roi_target(self, pos=None, size=None):
        """
        Move and scale the target ROI (used to display a particular area,
        for instance the currently scanned points
        during a scan
        Parameters
        ----------
        pos: (iterable) setting the central position of the ROI in the view
        size: (iterable) setting the size of the ROI
        """
        if isinstance(self.roi_target, pgROI):
            if size is not None:
                x_offset, x_scaling, y_offset, y_scaling = self._get_axis_scaling_offset()
                size = list(np.divide(list(size), [x_scaling, y_scaling]))
                if list(self.roi_target.size()) != size:
                    self.roi_target.setSize(size, center=(0.5, 0.5))

            if pos is not None:
                pos = self.unscale_axis(*list(pos))
                pos = list(pos)
                if list(self.roi_target.pos()) != pos:
                    self.roi_target.setPos(pos)

        else:
            self.roi_target.set_crosshair_position(*list(pos))

    def setup_widgets(self):
        vertical_layout = QtWidgets.QVBoxLayout()
        vertical_layout.setContentsMargins(5, 5, 5, 5)
        self.parent_widget.setLayout(vertical_layout)
        splitter_vertical = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        vertical_layout.addWidget(splitter_vertical)
        splitter_vertical.addWidget(self.toolbar)

        self.graphs_widget = QtWidgets.QWidget()
        self.graphs_widget.setLayout(QtWidgets.QHBoxLayout())
        self.graphs_widget.layout().setContentsMargins(0, 0, 0, 0)
        self.setup_graphs(self.graphs_widget.layout())
        splitter_vertical.addWidget(self.graphs_widget)

        self.plotitem.addItem(self.ROIselect)

        self.splitter_VLeft.splitterMoved[int, int].connect(self.move_right_splitter)
        self.splitter_VRight.splitterMoved[int, int].connect(self.move_left_splitter)

        self.splitter_VLeft.setSizes([1, 0])
        self.splitter_VRight.setSizes([1, 0])

    def setup_graphs(self, graphs_layout):
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        graphs_layout.addWidget(self.splitter)

        self.widget_histo = QtWidgets.QWidget()
        graphs_layout.addWidget(self.widget_histo)
        self.widget_histo.setLayout(QtWidgets.QHBoxLayout())

        self.splitter_VLeft = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self.splitter_VRight = QtWidgets.QSplitter(QtCore.Qt.Vertical)

        self.splitter.addWidget(self.splitter_VLeft)
        self.splitter.addWidget(self.splitter_VRight)

        self.splitter_VLeft.addWidget(self.image_widget)
        self.splitter_VLeft.addWidget(self._lineout_widgets['hor'])
        self.splitter_VRight.addWidget(self._lineout_widgets['ver'])
        self.splitter_VRight.addWidget(self._lineout_widgets['int'])

        self.image_widget.add_scaled_axis('right')
        self.image_widget.add_scaled_axis('top')
        self.image_widget.add_scaled_axis('left')
        self.image_widget.add_scaled_axis('bottom')
        self.splitter.addWidget(self.roi_manager.roiwidget)
        self.roi_manager.roiwidget.setVisible(False)

    def setup_actions(self):

        self.add_action('position', '(,)')
        self.add_action('red', 'Red Channel', 'r_icon', tip='Show/Hide Red Channel', checkable=True)
        self.add_action('green', 'Green Channel', 'g_icon', tip='Show/Hide Green Channel', checkable=True)
        self.add_action('blue', 'Blue Channel', 'b_icon', tip='Show/Hide Blue Channel', checkable=True)
        self.get_action('red').setChecked(True)
        self.get_action('green').setChecked(True)
        self.get_action('blue').setChecked(True)

        self.add_action('autolevels', 'AutoLevels', 'autoscale',
                        tip='Scale Histogram to Min/Max intensity', checkable=True)
        self.add_action('auto_levels_sym', 'AutoLevels Sym.', 'autoscale',
                        tip='Make the autoscale of the histograms symetric with respect to 0', checkable=True)

        self.add_action('histo', 'Histogram', 'Histogram', tip='Show/Hide Histogram', checkable=True)
        self.add_action('roi', 'ROI', 'Region', tip='Show/Hide ROI Manager', checkable=True)
        self.add_action('isocurve', 'IsoCurve', 'meshPlot', tip='Show/Hide Isocurve', checkable=True)
        self.add_action('aspect_ratio', 'Aspect Ratio', 'Zoom_1_1', tip='Fix Aspect Ratio', checkable=True, checked=True)
        self.add_action('crosshair', 'CrossHair', 'reset', tip='Show/Hide data Crosshair', checkable=True)
        self.add_action('ROIselect', 'ROI Select', 'Select_24',
                        tip='Show/Hide ROI selection area', checkable=True)
        self.add_action('flip_ud', 'Flip UD', 'scale_vertically',
                        tip='Flip the image up/down', checkable=True)
        self.add_action('flip_lr', 'Flip LR', 'scale_horizontally',
                        tip='Flip the image left/right', checkable=True)
        self.add_action('rotate', 'Rotate', 'rotation2',
                        tip='Rotate the image', checkable=True)
        self.add_action('legend', 'Legend', 'RGB',
                        tip='Show the legend', checkable=True)

    def update_colors(self, colors: list):
        for ind, roi_name in enumerate(self.roi_manager.ROIs):
            self.lineout_viewers['hor'].update_colors(make_dashed_pens(colors[ind]), displayer=roi_name)
            self.lineout_viewers['ver'].update_colors(make_dashed_pens(colors[ind]), displayer=roi_name)
            self.lineout_viewers['int'].update_colors(make_dashed_pens(colors[ind]), displayer=roi_name)

    def connect_things(self):
        self.connect_action('histo', self.histogrammer.activated)
        self.connect_action('autolevels', self.histogrammer.set_autolevels)
        self.roi_manager.new_ROI_signal.connect(self.update_roi_channels)
        self.roi_manager.new_ROI_signal.connect(self.add_roi_displayer)
        self.roi_manager.new_ROI_signal.connect(self.lineout_viewers['int'].get_action('clear').click)
        self.roi_manager.remove_ROI_signal.connect(self.remove_roi_displayer)
        self.roi_manager.color_signal.connect(self.update_colors)
        self.connect_action('isocurve', self.get_action('histo').trigger)

        self.connect_action('aspect_ratio', self.lock_aspect_ratio)
        self.connect_action('histo', self.show_hide_histogram)
        self.connect_action('roi', self.show_lineout_widgets)
        self.connect_action('roi', self.roi_clicked)
        self.connect_action('ROIselect', self.show_ROI_select)
        self.connect_action('crosshair', self.show_hide_crosshair)
        self.connect_action('crosshair', self.show_lineout_widgets)
        self.connect_action('legend', self.show_legend)

    def show_legend(self, show=True):
        self.data_displayer.show_legend(show)

    @Slot(int, str, str)
    def add_roi_displayer(self, index, roi_type='', roi_name=''):
        color = self.roi_manager.ROIs[roi_name].color
        self.lineout_viewers['hor'].view.add_data_displayer(roi_name, make_dashed_pens(color))
        self.lineout_viewers['ver'].view.add_data_displayer(roi_name, make_dashed_pens(color))
        self.lineout_viewers['int'].view.add_data_displayer(roi_name, make_dashed_pens(color))

    @Slot(str)
    def remove_roi_displayer(self, roi_name=''):
        self.lineout_viewers['hor'].view.remove_data_displayer(roi_name)
        self.lineout_viewers['ver'].view.remove_data_displayer(roi_name)
        self.lineout_viewers['int'].view.remove_data_displayer(roi_name)

    @Slot(int, str, str)
    def update_roi_channels(self, index, roi_type=''):
        """Update the use_channel setting each time a ROI is added"""
        self.roi_manager.update_use_channel(self.data_displayer.labels.copy())

    def prepare_ui(self):
        self.ROIselect.setVisible(False)
        self.show_hide_crosshair(False)
        self.show_lineout_widgets()

    @Slot(DataRaw)
    def display_images(self, datas):
        self.data_displayer.update_data(datas)
        if self.is_action_checked('isocurve'):
            self.isocurver.set_isocurve_data(datas.data[0])

    def display_roi_lineouts(self, roi_dte: DataToExport):
        if len(roi_dte) > 0:
            for lineout_type in self.lineout_types:
                for displayer_name in self.lineout_viewers[lineout_type].view.other_data_displayers:
                    dwa = roi_dte.get_data_from_name_origin(lineout_type, displayer_name)
                    if dwa is not None:
                        self.lineout_viewers[lineout_type].view.display_data(dwa.deepcopy(),
                                                                             displayer=displayer_name)

    def display_crosshair_lineouts(self, crosshair_dte: DataToExport):
        for lineout_type in self.lineout_types:
            dwa = crosshair_dte.get_data_from_name(lineout_type)
            if dwa is not None:
                self.lineout_viewers[lineout_type].view.display_data(dwa, displayer='crosshair')

    def show_lineout_widgets(self):
        state = self.is_action_checked('roi') or self.is_action_checked('crosshair')
        if state:
            self.prepare_image_widget_for_lineouts()
        else:
            self.prepare_image_widget_for_lineouts(1)

    @Slot(bool)
    def roi_clicked(self, isroichecked=True):
        self.roi_manager.roiwidget.setVisible(isroichecked)

        for k, roi in self.roi_manager.ROIs.items():
            roi.setVisible(isroichecked)

    def get_visible_images(self):
        are_items_visible = []
        for key in IMAGE_TYPES:
            are_items_visible.append(self.is_action_visible(key) and self.is_action_checked(key))
        return are_items_visible

    def notify_visibility_data_displayer(self):
        are_items_visible = self.get_visible_images()
        self.data_displayer.update_image_visibility(are_items_visible)

    @Slot(bool)
    def show_hide_histogram(self, show=True):
        are_items_visible = self.get_visible_images()
        self.histogrammer.show_hide_histogram(show, are_items_visible)

    def prepare_image_widget_for_lineouts(self, ratio=0.7):
        QtGui.QGuiApplication.processEvents()
        self.splitter_VRight.splitterMoved[int, int].emit(int(ratio * self.parent_widget.height()), 1)
        self.splitter.moveSplitter(int(ratio * self.parent_widget.width()), 1)
        self.splitter_VLeft.moveSplitter(int(ratio * self.parent_widget.height()), 1)
        self.splitter_VLeft.splitterMoved[int, int].emit(int(ratio * self.parent_widget.height()), 1)
        QtGui.QGuiApplication.processEvents()

    def collapse_lineout_widgets(self):
        self.prepare_image_widget_for_lineouts(ratio=1)

    def get_view_range(self):
        return self.image_widget.view.viewRange()

    def get_data_at(self, name='red', xy=(0, 0)):
        return self.data_displayer.get_image(name).get_val_at(xy)

    def lock_aspect_ratio(self):
        lock = self.is_action_checked('aspect_ratio')
        self.plotitem.vb.setAspectLocked(lock=lock, ratio=1)

    @Slot(int, int)
    def move_left_splitter(self, pos, index):
        self.splitter_VLeft.blockSignals(True)
        self.splitter_VLeft.moveSplitter(pos, index)
        self.splitter_VLeft.blockSignals(False)

    @Slot(int, int)
    def move_right_splitter(self, pos, index):
        self.splitter_VRight.blockSignals(True)
        self.splitter_VRight.moveSplitter(pos, index)
        self.splitter_VRight.blockSignals(False)

    def get_double_clicked(self):
        return self.image_widget.view.sig_double_clicked

    def get_axis(self, position='left') -> AxisItem_Scaled:
        if position not in AXIS_POSITIONS:
            raise KeyError(f'{position} is not a possible position for Axis: {AXIS_POSITIONS}')
        return self.image_widget.getAxis(position)

    @property
    def plotitem(self):
        return self.image_widget.plotitem

    def get_crosshair_signal(self):
        """Convenience function from the Crosshair"""
        return self.crosshair.crosshair_dragged

    def get_crosshair_position(self):
        """Convenience function from the Crosshair"""
        return self.crosshair.get_positions()

    def set_crosshair_position(self, *positions):
        """Convenience function from the Crosshair"""
        self.crosshair.set_crosshair_position(*positions)

    @Slot(bool)
    def show_hide_crosshair(self, show=True):
        self.crosshair.setVisible(show)
        self.set_action_visible('position', show)
        self.crosshair.setVisible(show)
        if show:
            self.lineout_viewers['hor'].view.add_data_displayer('crosshair', plot_colors=crosshair_pens)
            self.lineout_viewers['ver'].view.add_data_displayer('crosshair', plot_colors=crosshair_pens)
            self.lineout_viewers['int'].view.add_data_displayer('crosshair', plot_colors=crosshair_pens)
            range = self.get_view_range()
            self.set_crosshair_position(np.mean(np.array(range[0])), np.mean(np.array(range[0])))
        else:
            self.lineout_viewers['hor'].view.remove_data_displayer('crosshair')
            self.lineout_viewers['ver'].view.remove_data_displayer('crosshair')
            self.lineout_viewers['int'].view.remove_data_displayer('crosshair')
        logger.debug(f'Crosshair visible?: {self.crosshair.isVisible()}')

    def show_ROI_select(self):
        self.ROIselect.setVisible(self.is_action_checked('ROIselect'))
        rect = self.data_displayer.get_image('red').boundingRect()
        self.ROIselect.setPos(rect.center()-QtCore.QPointF(rect.width() * 2 / 3, rect.height() * 2 / 3)/2)
        self.ROIselect.setSize(rect.size() * 2 / 3)

    def set_image_labels(self, labels: List[str]):
        if self.data_displayer.labels != labels:
            action_names =['red', 'green', 'blue']
            for action_name, label in zip(action_names[:len(labels)], labels):
                self.get_action(action_name).setToolTip('Show/Hide'
                                                        f' - '
                                                        f'{label}')

    def set_axis_label(self, position, label='', units=''):
        """
        Convenience method to set label and unit of any view axes
        Parameters
        ----------
        position: (str) any of AXIS_POSITIONS
        label: (str) text of the axis label
        units: (str) units of the axis label
        """
        axis = self.get_axis(position)
        axis.setLabel(text=label, units=units)

    def get_axis_label(self, position):
        axis = self.get_axis(position)
        return axis.axis_label, axis.axis_units

    def set_axis_scaling(self, position='top', scaling=1, offset=0, label='', units='Pxls'):
        """
        Method used to update the scaling of the right and top axes in order to translate pixels to real coordinates
        Parameters
        ----------
        position: (str) axis position either one of AXIS_POSITIONS
        scaling: (float) scaling of the axis
        offset: (float) offset of the axis
        label: (str) text of the axis label
        units: (str) units of the axis label
        """
        if scaling is None:
            scaling = 1
        if offset is None:
            offset = 0
        self.get_axis(position).axis_scaling = scaling
        self.get_axis(position).axis_offset = offset
        self.set_axis_label(position, label=label, units=units)

    def scale_axis(self, xaxis, yaxis):
        """scale view coordinates from the regular axes to the scaled/offset ones"""
        x_offset, x_scaling, y_offset, y_scaling = self._get_axis_scaling_offset()
        return xaxis * x_scaling + x_offset, yaxis * y_scaling + y_offset

    def unscale_axis(self, xaxis, yaxis):
        """scale view coordinates from the scaled/offset axes to the regular ones"""
        x_offset, x_scaling, y_offset, y_scaling = self._get_axis_scaling_offset()
        return (xaxis - x_offset) / x_scaling, (yaxis - y_offset) / y_scaling

    def _get_axis_scaling_offset(self):
        x_offset = self.get_axis('top').axis_offset
        x_scaling = self.get_axis('top').axis_scaling
        y_offset = self.get_axis('right').axis_offset
        y_scaling = self.get_axis('right').axis_scaling
        return x_offset, x_scaling, y_offset, y_scaling


class Viewer2D(ViewerBase):
    """Object managing plotting and manipulation of 2D data using a View2D"""

    def __init__(self, parent: QtWidgets.QWidget = None, title=''):
        super().__init__(parent, title)

        self.just_init = True

        self._datas = None
        self.isdata = dict([])
        self._is_gradient_manually_set = False

        self.view = View2D(parent)
        self.filter_from_rois = Filter2DFromRois(self.view.roi_manager, self.view.data_displayer.get_image('red'),
                                                 IMAGE_TYPES)
        self.filter_from_rois.register_activation_signal(self.view.get_action('roi').triggered)
        self.filter_from_rois.register_target_slot(self.process_roi_lineouts)

        self.filter_from_crosshair = Filter2DFromCrosshair(self.view.crosshair, self.view.data_displayer.get_images(),
                                                           IMAGE_TYPES)
        self.filter_from_crosshair.register_activation_signal(self.view.get_action('crosshair').triggered)
        self.filter_from_crosshair.register_target_slot(self.process_crosshair_lineouts)

        self.prepare_connect_ui()

    @property
    def roi_manager(self):
        """Convenience method """
        return self.view.roi_manager

    @property
    def roi_target(self) -> pgROI:
        return self.view.roi_target

    def move_roi_target(self, pos: Iterable[float] = None, size: Iterable[float] = (1, 1)):
        """move a specific read only ROI at the given position on the viewer"""
        self.view.move_scale_roi_target(pos, size)

    @property
    def crosshair(self):
        """Convenience method """
        return self.view.crosshair

    @property
    def image_widget(self):
        """Convenience method """
        return self.view.image_widget

    def get_data_at(self):
        """Convenience method """
        return self.view.get_data_at()

    def set_crosshair_position(self, xpos, ypos):
        """Convenience method to set the crosshair positions"""
        self.view.crosshair.set_crosshair_position(xpos=xpos, ypos=ypos)

    def activate_roi(self, activate=True):
        """Activate the Roi manager using the corresponding action"""
        self.view.set_action_checked('roi', activate)
        self.view.get_action('roi').triggered.emit(activate)

    def roi_changed(self, *args, **kwargs):
        self.filter_from_rois.filter_data(self._datas)

    def crosshair_changed(self):
        self.filter_from_crosshair.filter_data(self._datas)

    def set_gradient(self, image_key, gradient):
        """convenience function"""
        self.view.histogrammer.set_gradient(image_key, gradient)

    def _show_data(self, data: DataWithAxes, *args, **kwargs):
        """Data to be plotted and eventually filtered using ROI...

        Parameters
        ----------
        data: DataWithAxes

        """

        if len(data) == 1 and not self._is_gradient_manually_set:
            self.set_gradient('red', 'grey')
        if len(data) > 3:
            logger.warning('Cannot plot on 2D plot more than 3 channels')
            data.data = data.data[:3]
        self.view.set_image_labels(data.labels)
        if data.distribution != self.view.data_displayer.display_type:
            self.view.set_image_displayer(data.distribution)
            self.filter_from_crosshair.set_graph_items(self.view.data_displayer.get_images())

        self.get_axes_from_view(data)  # in case axes were not specified into data, one try to get them from the view

        self.isdata['red'] = len(data) > 0
        self.isdata['green'] = len(data) > 1
        self.isdata['blue'] = len(data) > 2

        self.update_data()

        self.set_visible_items()
        if not self.view.is_action_checked('roi'):
            self.data_to_export_signal.emit(self.data_to_export)

        self.autolevels_first()

    def autolevels_first(self):
        if self.just_init and not self.is_action_checked('autolevels'):
            self.get_action('autolevels').trigger()
            self.update_data()
            self.get_action('autolevels').trigger()
            self.just_init = False

    def get_axes_from_view(self, data: DataWithAxes):
        """Obtain axes info from the view

        Only for uniform data
        """
        if data.distribution == DataDistribution['uniform']:
            if data.get_axis_from_index(0)[0] is None:
                axis_view = self.view.get_axis('right')
                axis = Axis(axis_view.axis_label, units=axis_view.axis_units,
                            scaling=axis_view.axis_scaling, offset=axis_view.axis_offset, index=0)
                axis.create_linear_data(data.shape[0])
                data.axes.append(axis)
            if data.get_axis_from_index(1)[0] is None:
                axis_view = self.view.get_axis('top')
                axis = Axis(axis_view.axis_label, units=axis_view.axis_units,
                            scaling=axis_view.axis_scaling, offset=axis_view.axis_offset, index=1)
                axis.create_linear_data(data.shape[1])
                data.axes.append(axis)

    def update_data(self):
        if self._raw_data is not None:
            self._datas = self.set_image_transform()
            if self._datas.distribution.name == 'uniform':
                self.x_axis = self._datas.get_axis_from_index(1)[0]
                self.y_axis = self._datas.get_axis_from_index(0)[0]
            else:
                self.x_axis = self._datas.get_axis_from_index(0)[0]
                self.y_axis = self._datas.get_axis_from_index(0)[1]
            self.view.display_images(self._datas)

            if self.view.is_action_checked('roi'):
                self.roi_changed()

            if self.view.is_action_checked('crosshair'):
                self.crosshair_changed()

    def set_image_transform(self) -> DataRaw:
        """
        Deactivate some tool buttons if data type is "spread" then apply transform_image
        """
        data = copy.deepcopy(self._raw_data)
        self.view.set_action_visible('flip_ud', data.distribution != 'spread')
        self.view.set_action_visible('flip_lr', data.distribution != 'spread')
        self.view.set_action_visible('rotate', data.distribution != 'spread')
        if data.distribution != 'spread':
            for ind_data in range(len(data)):
                data.data[ind_data] = self.transform_image(data.data[ind_data])
        return data

    def transform_image(self, data):
        if self.view.is_action_checked('flip_ud'):
            data = np.flipud(data)
        if self.view.is_action_checked('flip_lr'):
            data = np.fliplr(data)
        if self.view.is_action_checked('rotate'):
            data = np.flipud(np.transpose(data))
        return data

    def set_visible_items(self):
        for key in IMAGE_TYPES:
            self.view.set_action_visible(key, self.isdata[key])

        self.view.notify_visibility_data_displayer()

    def show_roi(self, show=True, show_roi_widget=True):
        """convenience function to control roi"""
        if show == (not self.view.is_action_checked('roi')):
            self.view.get_action('roi').trigger()

        self.view.roi_manager.roiwidget.setVisible(show_roi_widget)

    def update_crosshair_data(self, crosshair_dte: DataToExport):
        try:
            posx, posy = self.view.get_crosshair_position()
            (posx_scaled, posy_scaled) = self.view.scale_axis(posx, posy)

            dat = f'({posx_scaled:.1e}{posy_scaled:.1e})\n'
            dwa_int = crosshair_dte.get_data_from_name('int')
            if dwa_int is not None:
                for ind_data in range(len(dwa_int)):
                    dat += f' {dwa_int.labels[ind_data]}:{float(dwa_int[ind_data][0]):.1e}\n'

                self.view.set_action_text('position', dat)

        except Exception as e:
            logger.warning(str(e))

    def prepare_connect_ui(self):
        self.view.ROIselect.sigRegionChangeFinished.connect(self.selected_region_changed)

        self.roi_manager.roi_changed.connect(self.roi_changed)
        self.roi_manager.roi_value_changed.connect(self.roi_changed)

        self.view.connect_action('flip_ud', slot=self.update_data)
        self.view.connect_action('flip_lr', slot=self.update_data)
        self.view.connect_action('rotate', slot=self.update_data)
        self.view.connect_action('autolevels', slot=self.update_data)
        self.view.connect_action('isocurve', slot=self.update_data)
        self.view.histogrammer.gradient_changed.connect(lambda: setattr(self, '_is_gradient_manually_set', True))

        # todo : self.view.lineout_plotter.roi_changed.connect(self.roi_changed)
        self.view.get_crosshair_signal().connect(self.crosshair_changed)

        self.view.get_double_clicked().connect(self.double_clicked)

    def selected_region_changed(self):
        if self.view.is_action_checked('ROIselect'):
            pos = self.view.ROIselect.pos()
            size = self.view.ROIselect.size()
            # self.ROI_select_signal.emit(QtCore.QRectF(pos[0], pos[1], size[0], size[1]))
            self.roi_select_signal.emit(RoiInfo.info_from_rect_roi(self.view.ROIselect))

    @Slot(float, float)
    def double_clicked(self, posx, posy):
        if self.view.is_action_checked('crosshair'):
            self.view.crosshair.set_crosshair_position(posx, posy)
            self.crosshair_changed()
        #scale positions of double_click with respect to real axes
        posx, posy = self.view.scale_axis(posx, posy)
        self.sig_double_clicked.emit(posx, posy)


    @property
    def x_axis(self):
        return self.view.get_axis('top')

    @x_axis.setter
    def x_axis(self, axis: Axis = None):
        if axis is not None:
            self.view.set_axis_scaling('top', scaling=axis.scaling, offset=axis.offset,
                                       label=axis.label, units=axis.units)

    @property
    def y_axis(self):
        return self.view.get_axis('right')

    @y_axis.setter
    def y_axis(self, axis: Axis = None):
        if axis is not None:
            self.view.set_axis_scaling('right', scaling=axis.scaling, offset=axis.offset,
                                       label=axis.label, units=axis.units)

    @Slot(DataToExport)
    def process_crosshair_lineouts(self, dte):
        self.view.display_crosshair_lineouts(dte)
        self.update_crosshair_data(dte)
        self.crosshair_dragged.emit(*self.view.scale_axis(*self.view.crosshair.get_positions()))

    def process_roi_lineouts(self, roi_dte: DataToExport):
            if len(roi_dte) > 0:
                self.view.display_roi_lineouts(roi_dte)
                roi_dte_bis = roi_dte.deepcopy()
                for dwa in roi_dte_bis.data:
                    if dwa.name == 'hor':
                        dwa.name = f'Hlineout_{dwa.origin}'
                    elif dwa.name == 'ver':
                        dwa.name = f'Vlineout_{dwa.origin}'
                    elif dwa.name == 'int':
                        dwa.name = f'Integrated_{dwa.origin}'
                self.data_to_export.append(roi_dte_bis)

                self.measure_data_dict = dict([])

                for roi_name in roi_dte_bis.get_origins():
                    dwa = roi_dte_bis.get_data_from_name_origin(f'Integrated_{roi_name}', roi_name)
                    for ind, data_array in enumerate(dwa.data):
                            self.measure_data_dict[f'{dwa.labels[ind]}:'] = float(data_array[0])

                    QtWidgets.QApplication.processEvents()

                self.view.roi_manager.settings.child('measurements').setValue(self.measure_data_dict)
                if not self._display_temporary:
                    self.data_to_export_signal.emit(self.data_to_export)
                self.ROI_changed.emit()


def main_spread():
    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()
    prog = Viewer2D(widget)
    widget.show()

    def print_data(data: DataToExport):
        print(data)
        print('******')
        print(data.get_data_from_dim('Data1D'))


    prog.data_to_export_signal.connect(print_data)

    data_spread = np.load('../../../resources/triangulation_data.npy')

    prog.view.get_action('histo').trigger()
    prog.view.get_action('autolevels').trigger()

    prog.show_data(DataRaw(name='mydata', distribution='spread', data=[data_spread],
                           axes=[]))

    sys.exit(app.exec_())


def main(data_distribution='uniform'):
    """either 'uniform' or 'spread'"""

    app = QtWidgets.QApplication(sys.argv)
    widget = QtWidgets.QWidget()

    widget_button = QtWidgets.QWidget()
    widget_button.setLayout(QtWidgets.QHBoxLayout())
    button = QtWidgets.QPushButton('New Data')
    ndata = QtWidgets.QSpinBox()
    widget_button.layout().addWidget(button)
    widget_button.layout().addWidget(ndata)

    def print_data(data: DataToExport):
        print(data)
        print('******')
        print(data.get_data_from_dim('Data1D'))

    if data_distribution == 'uniform':
        data_to_plot = generate_uniform_data()

    elif data_distribution == 'spread':
        data_spread = np.load('../../../resources/triangulation_data.npy')
        data_to_plot = DataRaw(name='mydata', distribution='spread', data=[data_spread[:,2]],
                                       nav_indexes=(0,),
                                       axes=[Axis('xaxis', units='xpxl', data=data_spread[:,0], index=0, spread_order=0),
                                             Axis('yaxis', units='ypxl', data=data_spread[:,1], index=0, spread_order=1),])

    prog = Viewer2D(widget)
    widget.show()
    prog.data_to_export_signal.connect(print_data)

    prog.view.get_action('histo').trigger()
    prog.view.get_action('autolevels').trigger()

    prog.show_data(data_to_plot)

    prog.view.show_roi_target(True)
    prog.view.move_scale_roi_target((50, 40), (10, 20))

    button.clicked.connect(lambda: plot_data(prog, ndata.value()))
    widget_button.show()
    QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())


def generate_uniform_data() -> DataRaw:
    from pymodaq_utils.math_utils import gauss2D
    Nx = 100
    Ny = 2 * Nx
    data_random = np.random.normal(size=(Ny, Nx))
    x = 0.5 * np.linspace(-Nx / 2, Nx / 2 - 1, Nx)
    y = 0.2 * np.linspace(-Ny / 2, Ny / 2 - 1, Ny)
    data_red = 3 * np.sin(x / 5) ** 2 * gauss2D(x, 5, Nx / 10, y, -1, Ny / 10, 1, 90) + 0.2 * data_random
    data_green = 10 * gauss2D(x, -20, Nx / 10, y, -10, Ny / 20, 1, 0)
    data_green[70:80, 7:12] = np.nan

    data_to_plot = DataRaw(name='mydata', distribution='uniform',
                                   data=[data_red, data_green, data_red-data_green],
                                   labels = ['myreddata', 'mygreendata'],
                                   axes=[Axis('xaxis', units='xpxl', data=x, index=1),
                                         Axis('yaxis', units='ypxl', data=y, index=0), ])
    return data_to_plot


def plot_data(viewer2D: Viewer2D, ndata: int = 2):
    if ndata > 0:
        dwa = generate_uniform_data()
        dwa.data = dwa.data[0:ndata]
        viewer2D.show_data(dwa)


def print_roi_select(rect):
    print(rect)


def main_view():
    app = QtWidgets.QApplication(sys.argv)

    form = QtWidgets.QWidget()
    prog = View2D(form)
    form.show()
    sys.exit(app.exec_())


if __name__ == '__main__':  # pragma: no cover

    #main_view()
    main('uniform')
    #main('spread')



"""
S Weber  2020
Examples of custome parameter types derived from pyqtgraph
"""
import sys
from qtpy import QtWidgets, QtCore
from collections import OrderedDict

import pymodaq_gui.utils.widgets.table as table

from pymodaq_gui.managers.parameter_manager import ParameterManager


class ParameterEx(ParameterManager):
    params = [
        {'title': 'Groups:', 'name': 'groups', 'type': 'group', 'children': [
            {'title': 'A visible group:', 'name': 'agroup', 'type': 'group', 'children': []},
            {'title': 'An hidden group:', 'name': 'bgroup', 'type': 'group', 'children': [], 'visible': False},  # this
            # visible option is not available in usual pyqtgraph group
        ]},

        {'title': 'Numbers:', 'name': 'numbers', 'type': 'group', 'children': [
            {'title': 'Standard float', 'name': 'afloat', 'type': 'float', 'value': 20., 'min': 1.,
             'tip': 'displays this text as a tooltip'},
            {'title': 'Linear Slide float', 'name': 'linearslidefloat', 'type': 'slide', 'value': 50, 'default': 50,
             'min': 0,
             'max': 123, 'subtype': 'linear'},
            {'title': 'Linear int Slide', 'name': 'linearslideint', 'type': 'slide', 'value': 50, 'default': 50, 'step':1,
             'min': 0,
             'max': 123, 'subtype': 'linear', 'int': True},
            {'title': 'Linear Slide with suffix', 'name': 'linearslidewithsuffixandsiPrefix', 'type': 'slide', 'value': 50, 'default': 50,
             'min': 0,
             'max': 1e6, 'subtype': 'linear','suffix':'V','siPrefix':True},             
            {'title': 'Log Slide float', 'name': 'logslidefloat', 'type': 'slide', 'value': 50, 'default': 50,
             'min': 1e-5,
             'max': 1e5, 'subtype': 'log','suffix':'V','siPrefix':True},
        ]},

        {'title': 'Booleans:', 'name': 'booleans', 'type': 'group', 'children': [
            {'title': 'Standard bool', 'name': 'abool', 'type': 'bool', 'value': True},
            {'title': 'bool push', 'name': 'aboolpush', 'type': 'bool_push', 'value': True, 'label': 'action'},
            {'title': 'A led', 'name': 'aled', 'type': 'led', 'value': False, 'tip': 'a led you cannot toggle'},
            {'title': 'A led', 'name': 'anotherled', 'type': 'led_push', 'value': True, 'tip': 'a led you can toggle'},
        ]},

        {'title': 'DateTime:', 'name': 'datetimes', 'type': 'group', 'children': [
            {'title': 'Time:', 'name': 'atime', 'type': 'time', 'value': QtCore.QTime.currentTime()},
            {'title': 'Date:', 'name': 'adate', 'type': 'date', 'value': QtCore.QDate.currentDate(),
             'format': 'dd/MM/yyyy'},
            {'title': 'DateTime:', 'name': 'adatetime', 'type': 'date_time',
             'value': QtCore.QDateTime(QtCore.QDate.currentDate(), QtCore.QTime.currentTime()),
             'format': 'MM/dd/yyyy hh:mm', 'tip': 'displays this text as a tooltip'},
        ]},
        {'title': 'An action', 'name': 'action', 'type': 'action'},  # action whose displayed text corresponds to title

        {'title': 'Lists:', 'name': 'lists', 'type': 'group', 'children': [
            {'title': 'Standard list:', 'name': 'alist', 'type': 'list',
             'limits': ['a value', 'another one'], 'value': 'a value'},
            {'title': 'List with add:', 'name': 'anotherlist', 'type': 'list',
             'limits': ['a value', 'another one'], 'value': 'a value',
             'show_pb': True, 'tip': 'when using the "show_pb" option, displays a plus button to add elt to the list'},
            {'title': 'List defined from a dict:', 'name': 'dict_list', 'type': 'list',
             'limits': {'xaxis': 0, 'yaxis': [0, 1, 2]},
             'value': 'yaxis',
             'tip': 'Such a parameter display text that are keys of a dict while'
                                                        'values could be any object'
             },
        ]},
        {'title': 'Browsing files:', 'name': 'browser', 'type': 'group', 'children': [
            {'title': 'Look for a file:', 'name': 'afile', 'type': 'browsepath', 'value': '', 'filetype': True,
             'tip': 'If filetype is True select a file otherwise a directory'},
            {'title': 'Look for a dir:', 'name': 'adir', 'type': 'browsepath', 'value': '', 'filetype': False,
             'tip': 'If filetype is True select a file otherwise a directory'},

        ]},
        {'title': 'Selectable items:', 'name': 'itemss', 'type': 'group', 'children': [
            {'title': 'Selectable items', 'name': 'items', 'type': 'itemselect',
             'value': dict(all_items=['item1', 'item2', 'item3', 'item4', 'item5'],
                           selected=['item2']),
             'tip': 'Press Ctrl+click  to select items in any order'},
            {'title': 'Selectable items', 'name': 'itemsbis', 'type': 'itemselect',
             'value': dict(all_items=['item1', 'item2', 'item3'], selected=['item2']),
             'tip': 'If show_pb is True, user can add items to the list', 'show_pb': True,},
            {'title': 'Removable items', 'name': 'itemsbisbis', 'type': 'itemselect',
             'value': dict(all_items=['item1', 'item2', 'item3'], selected=['item2']),
             'tip': 'If show_mb is True, user can remove selected items from the list', 'show_mb': True,},            
            {'title': 'Checkable items', 'name': 'itemscheckable', 'type': 'itemselect',
             'value': dict(all_items=['item1', 'item2', 'item3'], selected=['item2']),
             'tip': 'If checkbox is True, user can select item by checking/unchecking items. Remove items is still used with standard selections',
             'show_pb': True, 'checkbox': True, 'show_mb': True,},                        
            {'title': 'Dragable items', 'name': 'itemsdragablecheckable', 'type': 'itemselect',
             'value': dict(all_items=['item1', 'item2', 'item3'], selected=['item2']),
             'tip': 'If dragdrop is True, user can drag or drop items inside the list', 'checkbox': True, 'dragdrop': True}, 
        ]},  
        {'title': 'Plain text:', 'name': 'texts', 'type': 'group', 'children': [
            {'title': 'Standard str', 'name': 'atte', 'type': 'str', 'value': 'this is a string you can edit'},
            {'title': 'Plain text', 'name': 'text', 'type': 'text', 'value': 'this is some text'},
            {'title': 'Plain text', 'name': 'textpb', 'type': 'text_pb', 'value': 'this is some text',
             'tip': 'If text_pb type is used, user can add text to the parameter'},
        ]},

        {'title': 'Tables:', 'name': 'tables', 'type': 'group', 'children': [
            {'title': 'Table widget', 'name': 'tablewidget', 'type': 'table', 'value':
                OrderedDict(key1='data1', key2=24), 'header': ['keys', 'limits'], 'height': 100},
            {'title': 'Table view', 'name': 'tabular_table', 'type': 'table_view',
             'delegate': table.SpinBoxDelegate, 'menu': True,
             'value': table.TableModel([[0.1, 0.2, 0.3]], ['value1', 'value2', 'value3']),
             'tip': 'The advantage of the Table model lies in its modularity.\n For concrete examples see the'
                    'TableModelTabular and the TableModelSequential custom models in the'
                    ' pymodaq.utils.scanner module'},
        ]},  # The advantage of the Table model lies in its modularity for concrete examples see the
        # TableModelTabular and the TableModelSequential custom models in the pymodaq.utils.scanner module
    ]

    def __init__(self):
        super().__init__()

    def value_changed(self, param):
        """
        """
        print(f'The parameter {param.name()} changed its value to {param.value()}')


def main():
    from pymodaq_gui.utils.utils import mkQApp

    app = mkQApp('Parameters')

    ptree = ParameterEx()
    ptree.settings_tree.show()
    ptree.settings.child('itemss', 'itemsbis').setValue(dict(all_items=['item1', 'item2', 'item3'],
                                                             selected=['item3']))

    ptree.settings.child('itemss', 'itemsbis').setValue(dict(all_items=['item1', 'item2', 'item3'],
                                                             selected=['item1', 'item3']))

    app.exec()


if __name__ == '__main__':
    main()
